from big_thing_py.common import *
from big_thing_py.core.request import *
from big_thing_py.core.mqtt_message import *
from big_thing_py.core.argument import *
from big_thing_py.core.service import *

from func_timeout import StoppableThread, FunctionTimedOut
import threading


class MXFunction(MXService):
    def __init__(
        self,
        func: Callable,
        tag_list: List[MXTag],
        return_type: MXType,
        name: str = '',
        energy: float = 0,
        desc: str = '',
        thing_name: str = '',
        middleware_name: str = '',
        arg_list: List[MXArgument] = [],
        exec_time: float = 0,
        timeout: float = 0,
        range_type: MXRangeType = MXRangeType.SINGLE,
    ) -> None:
        super().__init__(
            func=func,
            tag_list=tag_list,
            name=name,
            energy=energy,
            desc=desc,
            thing_name=thing_name,
            middleware_name=middleware_name,
        )

        self._return_type = return_type
        self._arg_list = arg_list
        self._exec_time = exec_time
        self._timeout = timeout
        # TODO: range_type will be removed from MXFunction
        self._range_type = range_type

        self._return_value = None
        self._running = False
        self._running_scenario_list: List[str] = []

        # Queue
        self._publish_queue: Queue = None

        if self._return_type in [MXType.UNDEFINED] or isinstance(self._return_type, str):
            raise MXValueError('return_type cannot be undefined or `str` type')
        if (not len(self._arg_list) == len(get_function_info(self._func)['args'])) if self._func else False:
            raise MXValueError('Length of argument list must be same with callback function')

        # TODO: return type detection feature
        # if not self._return_type or self._return_type in [MXType.UNDEFINED]:
        #     func_info = get_function_info(self._func)
        #     self._return_type = MXType().get(func_info[2])
        #     raise MXValueError('type must be specified')
        # elif self._return_type in [MXType.UNDEFINED]:
        #     raise MXValueError('type cannot be undefined or void')

    def __eq__(self, o: 'MXFunction') -> bool:
        instance_check = isinstance(o, MXFunction)
        arg_list_check = o._arg_list == self._arg_list
        return_type_check = o._return_type == self._return_type
        exec_time_check = o._exec_time == self._exec_time
        timeout_check = o._timeout == self._timeout
        range_type_check = o._range_type == self._range_type

        return (
            super().__eq__(o) and instance_check and arg_list_check and return_type_check and exec_time_check and timeout_check and range_type_check
        )

    def __getstate__(self):
        state = super().__getstate__()

        state['_return_type'] = self._return_type
        state['_arg_list'] = self._arg_list
        state['_exec_time'] = self._exec_time
        state['_timeout'] = self._timeout
        state['_range_type'] = self._range_type

        del state['_return_value']
        del state['_running']
        del state['_publish_queue']
        del state['_running_scenario_list']

        return state

    def __setstate__(self, state):
        super().__setstate__(state)

        self._type = state['_type']
        self._min = state['_min']
        self._max = state['_max']
        self._cycle = state['_cycle']
        self._format = state['_format']

        self._return_value = None
        self._running = False
        self._publish_queue = None
        self._running_scenario_list = []

    def _wrapper(self, execute_request: Union[MXExecuteRequest, MXInnerExecuteRequest]) -> bool:
        execute_msg = execute_request.trigger_msg
        MXLOG_DEBUG(f'[FUNC RUN] run {self._name} function by {execute_msg.scenario}', 'green')

        try:
            if not isinstance(execute_msg, MXExecuteMessage):
                raise Exception(f'[{get_current_function_name()}] Wrong ExecuteMessage type - execute_msg: {type(execute_msg)}')

            self._running = True
            error = MXErrorCode.NO_ERROR

            if self._timeout:
                # self._return_value = func_timeout(self._timeout, self._func, args=(*execute_msg.tuple_arguments(), ))
                current_thread = threading.current_thread()
                self._return_value = self._run_with_timeout(
                    timeout=self._timeout,
                    func=self._func,
                    name=current_thread.name,
                    user_data=dict(scenario=execute_msg.scenario),
                    args=(*execute_msg.tuple_arguments(),),
                )
            else:
                self._return_value = self._func(*execute_msg.tuple_arguments())
        except KeyboardInterrupt as e:
            # TODO: for wrapup main thread, but not test it yet
            print_error(e)
            MXLOG_DEBUG('Function execution exit by user', 'red')
            raise e
        except FunctionTimedOut as e:
            MXLOG_DEBUG(f'[FUNC TIMEOUT] function {self._name} by scenario {execute_msg.scenario} was timeout!!!', 'yellow')
            error = MXErrorCode.TIMEOUT
        except Exception as e:
            print_error(e)
            MXLOG_DEBUG(
                f'[FUNC FAIL] function {self._name} by scenario {execute_msg.scenario} is failed while executing!!!',
                'red',
            )
            error = MXErrorCode.FAIL
        else:
            error = MXErrorCode.NO_ERROR
        finally:
            if isinstance(execute_request, MXExecuteRequest):
                self._send_TM_RESULT_EXECUTE(scenario=execute_msg.scenario, error=error, request_ID=execute_msg.request_ID)
            elif isinstance(execute_request, MXInnerExecuteRequest):
                self._send_TM_RESULT_EXECUTE(scenario=execute_msg.scenario, error=error, is_inner=True, client_id=execute_msg.client_id)

            execute_request.timer_end()
            self._running = False
            self._running_scenario_list.remove(execute_msg.scenario)
            MXLOG_DEBUG(
                f'[FUNC END] function {self._name} end. -> return value : {self._return_value}, duration: {execute_request.duration():.4f} Sec',
                'green',
            )

    def start_execute_thread(self, execute_msg: MXExecuteMessage) -> MXThread:
        execute_protocol = MXProtocolType.get(execute_msg.topic)
        if execute_protocol == MXProtocolType.Base.MT_EXECUTE:
            execute_request = MXExecuteRequest(
                trigger_msg=execute_msg,
                result_msg=MXExecuteResultMessage(function=self, scenario=execute_msg.scenario, action_type=MXActionType.EXECUTE),
            )
        elif execute_protocol == MXProtocolType.Base.MT_IN_EXECUTE:
            execute_request = MXInnerExecuteRequest(
                trigger_msg=execute_msg,
                result_msg=MXExecuteResultMessage(
                    function=self, scenario=execute_msg.scenario, client_id=execute_msg.client_id, action_type=MXActionType.INNER_EXECUTE
                ),
            )

        self._running_scenario_list.append(execute_msg.scenario)
        execute_request.timer_start()

        execute_thread = MXThread(
            target=self._wrapper,
            name=f'{self._func.__name__}_{execute_msg.scenario}_thread',
            daemon=True,
            args=(execute_request,),
        )
        execute_thread.start()

        return execute_thread

    def _run_with_timeout(
        self,
        timeout: float,
        func: Callable,
        name: str = '',
        user_data: dict = None,
        args: tuple = (),
        kwargs: dict = None,
    ):
        if not kwargs:
            kwargs = {}
        if not args:
            args = ()

        ret = []
        exception = []
        isStopped = False

        def funcwrap(args2, kwargs2):
            try:
                ret.append(func(*args2, **kwargs2))
            except FunctionTimedOut:
                # Don't print traceback to stderr if we time out
                pass
            except Exception as e:
                exc_info = sys.exc_info()
                if isStopped is False:
                    # Assemble the alternate traceback, excluding this function
                    #  from the trace (by going to next frame)
                    # Python3 reads native from __traceback__,
                    # python2 has a different form for "raise"
                    e.__traceback__ = exc_info[2].tb_next
                    exception.append(e)

        thread = StoppableThread(target=funcwrap, name=name, args=(args, kwargs))
        thread.daemon = True
        thread.user_data = user_data

        thread.start()
        thread.join(timeout)

        stopException = None
        if thread.is_alive():
            isStopped = True

            class FunctionTimedOutTempType(FunctionTimedOut):
                def __init__(self):
                    return FunctionTimedOut.__init__(self, '', timeout, func, args, kwargs)

            FunctionTimedOutTemp = type(
                'FunctionTimedOut' + str(hash("%d_%d_%d_%d" % (id(timeout), id(func), id(args), id(kwargs)))),
                FunctionTimedOutTempType.__bases__,
                dict(FunctionTimedOutTempType.__dict__),
            )

            stopException = FunctionTimedOutTemp
            thread._stopThread(stopException)
            thread.join(min(0.1, timeout / 50.0))
            raise FunctionTimedOut('', timeout, func, args, kwargs)
        else:
            # We can still cleanup the thread here..
            # Still give a timeout... just... cuz..
            thread.join(0.5)

        if exception:
            raise exception[0] from None

        if ret:
            return ret[0]

    def _send_TM_RESULT_EXECUTE(self, scenario: str, error: MXErrorCode, request_ID: str = None, is_inner: bool = False, client_id: str = '') -> None:
        if is_inner:
            action_type = MXActionType.INNER_EXECUTE
        else:
            action_type = MXActionType.EXECUTE
        execute_result_msg = self.generate_execute_result_message(
            scenario=scenario, error=error, request_ID=request_ID, action_type=action_type, client_id=client_id
        )
        execute_result_mqtt_msg = execute_result_msg.mqtt_message()
        self._publish_queue.put(execute_result_mqtt_msg)

    def dict(self) -> dict:
        return {
            "name": self._name,
            "description": self._desc,
            "exec_time": self._exec_time * 1000 if self._exec_time is not None else 0,
            "return_type": self._return_type.value,
            "energy": self._energy,
            "tags": [tag.dict() for tag in self._tag_list],
            "use_arg": 1 if self._arg_list else 0,
            "arguments": [argument.dict() for argument in self._arg_list] if self._arg_list else [],
        }

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    def generate_execute_result_message(
        self, scenario: str, error: MXErrorCode, request_ID: str, action_type: MXActionType = MXActionType.EXECUTE, client_id: str = ''
    ) -> MXExecuteResultMessage:
        execute_result_msg = MXExecuteResultMessage(
            function=self, scenario=scenario, request_ID=request_ID, error=error, action_type=action_type, client_id=client_id
        )
        return execute_result_msg

    # ====================================
    #               _    _
    #              | |  | |
    #   __ _   ___ | |_ | |_   ___  _ __
    #  / _` | / _ \| __|| __| / _ \| '__|
    # | (_| ||  __/| |_ | |_ |  __/| |
    #  \__, | \___| \__| \__| \___||_|
    #   __/ |
    #  |___/
    # ====================================

    def get_exec_time(self) -> float:
        return self._exec_time

    def get_timeout(self) -> float:
        return self._timeout

    def get_arg_list(self) -> List[MXArgument]:
        return self._arg_list

    def get_return_type(self) -> MXType:
        return self._return_type

    def get_return_value(self) -> Union[int, float, str, bool]:
        return self._return_value

    def get_running(self) -> bool:
        return self._running

    def get_range_type(self) -> MXRangeType:
        return self._range_type

    def get_running_scenario_list(self) -> List[str]:
        return self._running_scenario_list

    # ==================================
    #             _    _
    #            | |  | |
    #  ___   ___ | |_ | |_   ___  _ __
    # / __| / _ \| __|| __| / _ \| '__|
    # \__ \|  __/| |_ | |_ |  __/| |
    # |___/ \___| \__| \__| \___||_|
    # ==================================

    def set_exec_time(self, exec_time: float) -> None:
        self._exec_time = exec_time

    def set_timeout(self, timeout: float) -> None:
        self._timeout = timeout

    def set_arg_list(self, arg_list: List[MXArgument]) -> None:
        self._arg_list = arg_list

    def set_return_type(self, return_type: MXType) -> None:
        self._return_type = return_type

    def set_return_value(self, return_value: Union[int, float, str, bool]) -> None:
        self._return_value = return_value

    def set_running(self, running: bool) -> None:
        self._running = running

    def set_range_type(self, range_type: MXRangeType) -> None:
        self._range_type = range_type

    # for link to big_thing's publish_queue
    def set_publish_queue(self, queue: Queue):
        self._publish_queue = queue


if __name__ == '__main__':
    pass



from big_thing_py.utils import *


def make_super_request_key(scenario_name: str, requester_middleware_name: str) -> str:
    return '@'.join([scenario_name, requester_middleware_name])


def make_sub_service_request_key(sub_service_name: str, sub_service_request_order: int) -> str:
    return '@'.join([sub_service_name, str(sub_service_request_order)])


def make_request_ID(requester_middleware_name: str, super_thing_name: str, super_service_name: str, sub_service_request_order: int):
    return '@'.join([requester_middleware_name, super_thing_name, super_service_name, str(sub_service_request_order)])


class MXMQTTMessage:
    def __init__(self) -> None:
        self.protocol_type: Union[MXProtocolType.Base, MXProtocolType.Super, MXProtocolType.WebClient] = None
        self.timestamp: float = None
        self.topic: str = None

    def set_timestamp(self, timestamp: float = None) -> None:
        if timestamp:
            self.timestamp = timestamp
        else:
            self.timestamp = time.time()


class MXMQTTSendMessage(MXMQTTMessage):
    def __init__(self) -> None:
        super().__init__()
        self.payload: dict = None

    def mqtt_message(self) -> mqtt.MQTTMessage:
        msg = encode_MQTT_message(self.topic, self.payload)
        return msg


class MXMQTTReceiveMessage(MXMQTTMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__()
        self.topic, self.payload, self.timestamp = decode_MQTT_message(msg)
        self.topic_error: bool = False
        self.payload_error: bool = False


class MXRegisterMessage(MXMQTTSendMessage):
    def __init__(self, thing) -> None:
        from big_thing_py.core.thing import MXThing

        thing: MXThing = thing

        super().__init__()
        self.protocol_type = MXProtocolType.Base.TM_REGISTER
        self.topic = self.protocol_type.value % (thing.get_name())
        self.payload = thing.dict()


##############################################################################################################################


class MXRegisterResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Base.MT_RESULT_REGISTER
        self.thing_name = self.topic.split('/')[3]
        self.middleware_name: str = self.payload['middleware_name']
        self.error: MXErrorCode = MXErrorCode.get(self.payload['error'])


class MXUnregisterMessage(MXMQTTSendMessage):
    def __init__(self, thing) -> None:
        from big_thing_py.core.thing import MXThing

        thing: MXThing = thing

        super().__init__()
        self.protocol_type = MXProtocolType.Base.TM_UNREGISTER
        self.topic = self.protocol_type.value % (thing.get_name())
        self.payload = EMPTY_JSON


class MXUnregisterResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Base.MT_RESULT_UNREGISTER
        self.thing_name: str = self.topic.split('/')[3]
        self.error: MXErrorCode = MXErrorCode.get(self.payload['error'])


class MXExecuteMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.get(msg.topic)
        self.middleware_name: str = ''
        self.request_ID: str = ''
        self.client_id: str = ''

        if self.protocol_type == MXProtocolType.Base.MT_EXECUTE:
            self.function_name: str = self.topic.split('/')[2]
            self.thing_name: str = self.topic.split('/')[3]

            if len(self.topic.split('/')) == 6:
                # MT/EXECUTE/[FunctionName]/[ThingName]/[TargetMiddlewareName]/[Request_ID] topic
                self.middleware_name: str = self.topic.split('/')[4]
                self.request_ID: str = self.topic.split('/')[5]
                if not self.middleware_name or not self.request_ID:
                    self.topic_error = True
            elif len(self.topic.split('/')) == 4:
                # MT/EXECUTE/[FunctionName]/[ThingName] topic
                self.middleware_name: str = ''
                self.request_ID: str = ''
            else:
                self.topic_error = True
        elif self.protocol_type == MXProtocolType.Base.MT_IN_EXECUTE:
            self.function_name: str = self.topic.split('/')[3]
            self.thing_name: str = self.topic.split('/')[4]
            self.middleware_name: str = ''
            self.request_ID: str = ''

        self.scenario: str = self.payload.get('scenario', '')
        self.arguments: List[dict] = self.payload.get('arguments', [])
        self.client_id: str = self.payload.get('client_id', '')

        if not self.scenario or self.arguments == None:
            self.payload_error = True
        elif not all([isinstance(arg.get('order', None), int) for arg in self.arguments]) or not all(
            [isinstance(arg.get('value', None), (int, float, str, bool)) for arg in self.arguments]
        ):
            self.payload_error = True

    def tuple_arguments(self) -> tuple:
        self.arguments = sorted(self.arguments, key=lambda x: int(x['order']))
        real_arguments = tuple([argument['value'] for argument in self.arguments])
        return real_arguments

    def dict_arguments(self) -> List[dict]:
        self.arguments = sorted(self.arguments, key=lambda x: int(x['order']))
        json_arguments = [dict(order=arg['order'], value=arg['value']) for arg in self.arguments]
        return json_arguments


class MXExecuteResultMessage(MXMQTTSendMessage):
    def __init__(
        self,
        function,
        scenario: str,
        client_id: str = '',
        request_ID: str = '',
        error: MXErrorCode = MXErrorCode.UNDEFINED,
        action_type: MXActionType = MXActionType.EXECUTE,
    ) -> None:
        from big_thing_py.core.function import MXFunction

        super().__init__()

        if action_type == MXActionType.EXECUTE:
            self.protocol_type = MXProtocolType.Base.TM_RESULT_EXECUTE
        elif action_type == MXActionType.INNER_EXECUTE:
            self.protocol_type = MXProtocolType.Base.TM_IN_RESULT_EXECUTE

        self.function: MXFunction = function
        self.request_ID = request_ID
        self.scenario = scenario
        self.client_id = client_id
        self._error = error

        if self.request_ID:
            self.topic = self.protocol_type.value % (
                self.function.get_name(),
                self.function.get_thing_name(),
                self.function.get_middleware_name(),
                self.request_ID,
            )
        else:
            if action_type == MXActionType.EXECUTE:
                self.topic = (self.protocol_type.value % (self.function.get_name(), self.function.get_thing_name(), '', '')).rstrip('/')
            elif action_type == MXActionType.INNER_EXECUTE:
                self.topic = self.protocol_type.value % (self.function.get_name(), self.function.get_thing_name())

        if action_type == MXActionType.EXECUTE:
            self.payload = dict(
                error=self._error.value,
                scenario=self.scenario,
                return_type=self.function.get_return_type().value,
                return_value=self.function.get_return_value(),
            )
        elif action_type == MXActionType.INNER_EXECUTE:
            self.payload = dict(
                error=self._error.value,
                scenario=self.scenario,
                client_id=self.client_id,
                return_type=self.function.get_return_type().value,
                return_value=self.function.get_return_value(),
            )

    @property
    def error(self) -> dict:
        return self._error

    @error.setter
    def error(self, error: MXErrorCode):
        if not isinstance(error, MXErrorCode):
            raise MXTypeError("error must be an MXErrorType")
        self._error = error
        self.payload = dict(
            error=self._error.value,
            scenario=self.scenario,
            return_type=self.function.get_return_type().value,
            return_value=self.function.get_return_value(),
        )


# TODO: binary 부분 구현 완료 후 작성하기
class MXBinaryValueResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.Base.MT_RESULT_BINARY_VALUE
        self.thing_name = self.topic.split('/')[3]
        self.value_name = self.payload['value_name']


class MXAliveMessage(MXMQTTSendMessage):
    def __init__(self, thing) -> None:
        from big_thing_py.core.thing import MXThing

        thing: MXThing = thing

        super().__init__()
        self.protocol_type = MXProtocolType.Base.TM_ALIVE
        self.topic = self.protocol_type.value % (thing.get_name())
        self.payload = EMPTY_JSON


class MXValuePublishMessage(MXMQTTSendMessage):
    def __init__(self, thing, value) -> None:
        from big_thing_py.core.thing import MXThing
        from big_thing_py.core.value import MXValue

        thing: MXThing = thing
        value: MXValue = value

        super().__init__()
        self.protocol_type = MXProtocolType.Base.TM_VALUE_PUBLISH
        self.topic = self.protocol_type.value % (value.get_name(), thing.get_name())
        self.payload = value.publish_dict()


# for middleware detect
class MXClientServiceListResultMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.WebClient.ME_RESULT_SERVICE_LIST
        self._client_id = self.topic.split('/')[3]


class MXClientRefreshMessage(MXMQTTSendMessage):
    def __init__(self, thing) -> None:
        from big_thing_py.core.thing import MXThing

        thing: MXThing = thing

        super().__init__()
        self.protocol_type = MXProtocolType.WebClient.EM_REFRESH
        self.topic = self.protocol_type.value % (thing.get_name())
        self.payload = EMPTY_JSON


class MXNotifyMessage(MXMQTTReceiveMessage):
    def __init__(self, msg: mqtt.MQTTMessage) -> None:
        super().__init__(msg=msg)
        self.protocol_type = MXProtocolType.WebClient.ME_NOTIFY_CHANGE

        # topic
        self._client_id = self.topic.split('/')[2]

        # payload


if __name__ == '__main__':
    payload = {
        'scenario': 'test',
        'arguments': [
            {
                'order': 0,
                'value': 1,
            },
            {
                'order': 1,
                'value': 2,
            },
        ],
    }
    msg = encode_MQTT_message('MT/EXECUTE/test1/test2', payload)
    message = MXExecuteMessage(msg)
    print(message.tuple_arguments())
    print(message.dict_arguments(*(1, 2)))



from big_thing_py.common import *

import inspect
import socket
import json
import time
import platform
import os
import copy
import random
from pathlib import Path
import getmac
import re
import sys
import typing
import pkg_resources


def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func

    return decorate


def get_type_name(typ: typing.Type) -> str:
    origin = typing.get_origin(typ)
    if origin is None:
        return typ.__name__
    args = typing.get_args(typ)
    args_str = ", ".join(get_type_name(arg) for arg in args)
    return f"{origin.__name__}[{args_str}]"


def get_function_info(func: Callable) -> Dict[str, Union[str, List[Tuple[str, Any]]]]:
    sig = inspect.signature(func)
    params = [(param.name, param.annotation) for param in sig.parameters.values()]
    ret_type = sig.return_annotation
    if not ret_type is None:
        if isinstance(ret_type, str):
            try:
                ret_type = eval(ret_type)
            except NameError:
                pass
        ret_type = get_type_name(ret_type)
    return dict(name=func.__name__, args=params, return_type=ret_type)


def get_current_function_name() -> str:
    return inspect.stack()[1].function


def get_upper_function_name(step: int = 1, frame=None) -> str:
    if frame is None:
        frame = inspect.stack()[1].frame
    if step == 1:
        co_name = frame.f_back.f_code.co_name
    elif step > 1:
        co_name = get_upper_function_name(step - 1, frame.f_back)
    else:
        co_name = 'too many steps... return MAX upper function name'
    return co_name


def get_mac_address(interface: str = None) -> str:
    mac_address = getmac.get_mac_address(interface)

    if mac_address:
        mac_address = mac_address.replace(':', '').upper()
        return mac_address
    else:
        return None


def get_current_time(mode: TimeFormat = TimeFormat.UNIXTIME) -> Union[str, float]:
    if mode == TimeFormat.DATETIME1:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    elif mode == TimeFormat.DATETIME2:
        return time.strftime("%Y%m%d_%H%M%S", time.localtime())
    elif mode == TimeFormat.DATE:
        return time.strftime("%Y-%m-%d", time.localtime())
    elif mode == TimeFormat.TIME:
        return time.strftime("%H:%M:%S", time.localtime())
    elif mode == TimeFormat.UNIXTIME:
        return time.time()
    return time.time()


def type_converter(in_type: Union[MXType, type, str, None]) -> Union[type, MXType]:
    if type(in_type) == MXType:
        if in_type in (MXType.BINARY, MXType.STRING):
            return str
        elif in_type == MXType.BOOL:
            return bool
        elif in_type == MXType.DOUBLE:
            return float
        elif in_type == MXType.INTEGER:
            return int
        elif in_type in (MXType.VOID, MXType.UNDEFINED):
            return None
        else:
            raise MXNotSupportedError('Unexpected python type!!!')
    elif type(in_type) == type:
        if in_type == int:
            return MXType.INTEGER
        elif in_type == float:
            return MXType.DOUBLE
        elif in_type == bool:
            return MXType.BOOL
        elif in_type == bytes:
            return MXType.BINARY
        elif in_type == str:
            return MXType.STRING
        elif in_type == type(None):
            return MXType.VOID
        else:
            raise MXNotSupportedError('Unexpected MXType type!!!')
    elif type(in_type) == str:
        if in_type == 'int':
            return MXType.INTEGER
        elif in_type == 'void':
            return MXType.VOID
        elif in_type == 'double':
            return MXType.DOUBLE
        elif in_type == 'bool':
            return MXType.BOOL
        elif in_type == 'binary':
            return MXType.BINARY
        elif in_type == 'string':
            return MXType.STRING
        elif in_type == 'undefined':
            return MXType.UNDEFINED
        else:
            raise MXNotSupportedError('Unexpected MXType type!!!')
    elif in_type == None or type(in_type) == type(None):
        return MXType.VOID


def json_file_read(path: str) -> Union[dict, bool]:
    try:
        with open(path, 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        return False


def json_file_write(path: str, data: Union[str, dict], indent: int = 4) -> None:
    with open(path, 'w') as f:
        if isinstance(data, (dict, str)):
            json.dump(data, f, indent=indent)
        else:
            raise Exception(f'common_util.json_file_write: data type error - {type(data)}')


def get_project_root(project_name: str = 'big-thing-py') -> Path:
    start_path = Path(__file__)
    while True:
        if str(start_path).split('/')[-1] == project_name:
            return str(start_path)
        else:
            start_path = start_path.parent


def check_valid_identifier(identifier: str) -> bool:
    pattern = r'^[a-zA-Z0-9_]*$'
    return bool(re.match(pattern, identifier))


def convert_special_char(char: str) -> str:
    # conversion_dict = {
    #     ' ': 'space',
    #     '!': 'exclamation_mark',
    #     '"': 'double_quote',
    #     '#': 'hash',
    #     '$': 'dollar',
    #     '%': 'percent',
    #     '&': 'ampersand',
    #     "'": 'single_quote',
    #     '(': 'left_parenthesis',
    #     ')': 'right_parenthesis',
    #     '*': 'asterisk',
    #     '+': 'plus',
    #     ',': 'comma',
    #     '-': 'hyphen',
    #     '.': 'period',
    #     '/': 'slash',
    #     ':': 'colon',
    #     ';': 'semicolon',
    #     '<': 'less_than',
    #     '=': 'equals',
    #     '>': 'greater_than',
    #     '?': 'question_mark',
    #     '@': 'at_sign',
    #     '[': 'left_bracket',
    #     '\\': 'backslash',
    #     ']': 'right_bracket',
    #     '^': 'caret',
    #     '`': 'backtick',
    #     '{': 'left_brace',
    #     '|': 'vertical_bar',
    #     '}': 'right_brace',
    #     '~': 'tilde'
    # }
    conversion_dict = {
        ' ': 'sp',
        '!': 'ex',
        '"': 'dq',
        '#': 'hs',
        '$': 'dl',
        '%': 'pc',
        '&': 'am',
        "'": 'sq',
        '(': 'lp',
        ')': 'rp',
        '*': 'as',
        '+': 'pl',
        ',': 'cm',
        '-': 'mn',
        '.': 'pd',
        '/': 'sl',
        ':': 'cl',
        ';': 'sc',
        '<': 'lt',
        '=': 'eq',
        '>': 'gt',
        '?': 'qm',
        '@': 'at',
        '[': 'lb',
        '\\': 'bs',
        ']': 'rb',
        '^': 'cr',
        '`': 'bt',
        '{': 'lc',
        '|': 'vb',
        '}': 'rc',
        '~': 'td',
    }
    if char in conversion_dict:
        return f'_{conversion_dict[char]}_'
    else:
        return char


def convert_to_valid_string(input_string: str) -> str:
    converted_chars = [convert_special_char(char) for char in input_string]
    return ''.join(converted_chars)


def convert_url_to_ip(url: str) -> str:
    ip_list = url.split(".")
    not_ip = False

    for ip_token in ip_list:
        if not ip_token.isdigit():
            not_ip = True
            break

    if not_ip:
        host_ip = socket.gethostbyname(url.strip())
        return host_ip
    elif not all([(0 <= int(ip_token) <= 255) for ip_token in ip_list]) or len(ip_list) != 4:
        raise Exception('wrong ip format')
    else:
        return url


def check_python_version():
    if sys.version_info[0] <= 3 and sys.version_info[1] < 6:
        raise Exception('Avahi feature is not supported on python3.6<=. try python3.7 or higher. or disable avahi feature. (avahi_enable=False)')


def is_valid_ip_address(ip_address: str) -> bool:
    if ip_address:
        octets = ip_address.split('.')
    else:
        return False

    if len(octets) != 4:
        return False
    for octet in octets:
        try:
            value = int(octet)
        except ValueError:
            return False
        if not 0 <= value <= 255:
            return False
    return True


def is_raspberry_pi():
    try:
        with open('/proc/cpuinfo', 'r') as f:
            cpuinfo = f.read()
        return (platform.uname().system == 'Linux') and ('Raspberry Pi' in cpuinfo)
    except:
        return False


def check_os_architecture() -> str:
    architecture = platform.architecture()[0]
    return architecture


def install_missing_package(error):
    package_name = str(error).split("'")[1]
    os.system(f'pip install {package_name}')


def dummy_func(arg_list: list):
    args = ', '.join([str(arg) for arg in arg_list])
    lambda_str = f"lambda {args}: None"
    exec(lambda_str)
    return eval(lambda_str)


def sdk_version() -> Union[str, None]:
    try:
        return pkg_resources.get_distribution('big-thing-py').version
    except pkg_resources.DistributionNotFound:
        return None


if __name__ == '__main__':
    pass



from big_thing_py.big_thing import *
from big_thing_py.staff_thing import *
from big_thing_py.manager import *
import uuid


class MXPollManagerThing(MXBigThing, metaclass=ABCMeta):
    MANAGER_THREAD_TIME_OUT = THREAD_TIME_OUT

    def __init__(
        self,
        name: str,
        service_list: List[MXService],
        alive_cycle: float,
        desc: str = '',
        version: str = sdk_version(),
        is_super: bool = False,
        is_parallel: bool = True,
        ip: str = None,
        port: int = None,
        ssl_ca_path: str = None,
        ssl_cert_path: str = '',
        ssl_key_path: str = '',
        log_name: str = None,
        log_enable: bool = True,
        log_mode: MXPrintMode = MXPrintMode.ABBR,
        append_mac_address: bool = True,
        manager_mode: MXManagerMode = MXManagerMode.SPLIT,
        scan_cycle=5,
    ):
        super().__init__(
            name=name,
            desc=desc,
            version=version,
            service_list=service_list,
            alive_cycle=alive_cycle,
            is_super=is_super,
            is_parallel=is_parallel,
            ip=ip,
            port=port,
            ssl_ca_path=ssl_ca_path,
            ssl_cert_path=ssl_cert_path,
            ssl_key_path=ssl_key_path,
            log_name=log_name,
            log_enable=log_enable,
            log_mode=log_mode,
            append_mac_address=append_mac_address,
        )

        self._scan_cycle = scan_cycle
        self._manager_mode = MXManagerMode.get(manager_mode) if isinstance(manager_mode, str) else manager_mode

        self._staff_thing_list: List[MXStaffThing] = []
        self._manager_mode_handler = ManagerModeHandler(mode=self._manager_mode)

        self._last_scan_time = 0

        self._register_waiting_staff_thing_queue: Queue = Queue()
        self._unregister_waiting_staff_thing_queue: Queue = Queue()

        # Threading
        self._thread_func_list += [
            self._scan_staff_message_thread_func,
        ]

    @override
    def run(self):
        try:
            self._connect(self._ip, self._port)

            # Start main threads
            for thread in self._comm_thread_list + self._thread_list:
                thread.start()

            if self._manager_mode == MXManagerMode.JOIN:
                self._register(self)
            elif self._manager_mode == MXManagerMode.SPLIT:
                # If manager mode is SPLIT, manager thing will be hid itself from user.
                # This process run by the middleware.
                self._register(self)
            else:
                pass

            # Maintain main thread
            while not self._g_exit.wait(THREAD_TIME_OUT):
                time.sleep(1000)
        except KeyboardInterrupt as e:
            MXLOG_DEBUG('Ctrl + C Exit', 'red')
            return self.wrapup()
        except ConnectionRefusedError as e:
            MXLOG_DEBUG('Connection error while connect to broker. Check ip and port', 'red')
            return self.wrapup()
        except Exception as e:
            print_error(e)
            return self.wrapup()

    @override
    def wrapup(self):
        try:
            for staff_thing in self._staff_thing_list:
                self._send_TM_UNREGISTER(staff_thing)
            cur_time = get_current_time()

            self._g_exit.set()
            for thread in self._thread_list:
                thread.join()
                MXLOG_DEBUG(f'{thread._name} is terminated', 'yellow')

            while not ((self._publish_queue.empty() and self._receive_queue.empty()) or (get_current_time() - cur_time > 3)):
                time.sleep(THREAD_TIME_OUT)

            self._g_comm_exit.set()
            for thread in self._comm_thread_list:
                thread.join()
                MXLOG_DEBUG(f'{thread._name} is terminated', 'yellow')

            self._mqtt_client.disconnect()
            MXLOG_DEBUG('Thing Exit', 'red')
            return True
        except Exception as e:
            print_error(e)
            return False

    # ===========================================================================================
    #  _    _                             _    __                      _    _
    # | |  | |                           | |  / _|                    | |  (_)
    # | |_ | |__   _ __   ___   __ _   __| | | |_  _   _  _ __    ___ | |_  _   ___   _ __   ___
    # | __|| '_ \ | '__| / _ \ / _` | / _` | |  _|| | | || '_ \  / __|| __|| | / _ \ | '_ \ / __|
    # | |_ | | | || |   |  __/| (_| || (_| | | |  | |_| || | | || (__ | |_ | || (_) || | | |\__ \
    #  \__||_| |_||_|    \___| \__,_| \__,_| |_|   \__,_||_| |_| \___| \__||_| \___/ |_| |_||___/
    # ===========================================================================================

    @override
    def _alive_publishing_thread_func(self, stop_event: Event) -> Union[bool, None]:
        try:
            while not stop_event.wait(THREAD_TIME_OUT):
                if self._manager_mode == MXManagerMode.JOIN:
                    current_time = get_current_time()
                    if not current_time - self._last_alive_time > self._alive_cycle:
                        continue
                    for staff_thing in self._staff_thing_list:
                        self._send_TM_ALIVE(staff_thing)
                        staff_thing._last_alive_time = current_time
                elif self._manager_mode == MXManagerMode.SPLIT:
                    # check staff thing is alive
                    current_time = get_current_time()
                    for staff_thing in self._staff_thing_list:
                        if not current_time - staff_thing._last_alive_time > staff_thing._alive_cycle:
                            continue
                        self._send_TM_UNREGISTER(staff_thing)
                        staff_thing._is_connected = False
                        self._staff_thing_list.remove(staff_thing)
                    pass
                else:
                    raise Exception('Invalid Manager Mode')

                if get_current_time() - self._last_alive_time > self._alive_cycle:
                    self._send_TM_ALIVE(self)
        except Exception as e:
            stop_event.set()
            print_error(e)
            return False

    @override
    def _value_publishing_thread_func(self, stop_event: Event) -> Union[bool, None]:
        try:
            while not stop_event.wait(THREAD_TIME_OUT):
                for staff_thing in self._staff_thing_list:
                    if not staff_thing._is_registered:
                        continue
                    current_time = get_current_time()
                    for value in staff_thing._value_list:
                        if not (current_time - value.get_last_update_time()) > value.get_cycle():
                            continue

                        is_updated = value.update()
                        if is_updated is not None:
                            self._send_TM_VALUE_PUBLISH(staff_thing=staff_thing, value=value)
        except Exception as e:
            stop_event.set()
            print_error(e)
            return False

    ############################################################################################################################

    def _scan_staff_message_thread_func(self, stop_event: Event):
        '''
        scan staff thing periodically and create staff thing instance

        if scanned staff thing is not in self._staff_thing_list, create staff thing instance and
        connect to staff thing
        else if scanned staff thing is in self._staff_thing_list, send alive packet.
        '''
        try:
            while not stop_event.wait(self.MANAGER_THREAD_TIME_OUT):
                if not (get_current_time() - self._last_scan_time > self._scan_cycle):
                    continue

                scanned_staff_thing_info_list = self._scan_staff_thing()
                if scanned_staff_thing_info_list is False:
                    MXLOG_DEBUG('Scan staff thing failed.', 'red')
                    continue
                elif len(scanned_staff_thing_info_list) == 0:
                    MXLOG_DEBUG('No staff thing found.', 'yellow')
                    continue

                for staff_thing_info in scanned_staff_thing_info_list:
                    # create staff thing instance
                    staff_thing = self._create_staff(staff_thing_info)
                    if not staff_thing:
                        continue

                    # if scanned staff thing is already in self._staff_thing_list, send alive packet
                    # and update last_alive_time
                    exist_staff_thing = self._get_staff_thing_by_name(staff_thing.get_name())
                    if exist_staff_thing:
                        self._send_TM_ALIVE(exist_staff_thing)
                        continue

                    MXLOG_DEBUG(f'New staff_thing!!! name: [{staff_thing.get_name()}]', 'green')
                    staff_thing._is_connected = True

                    # else if scanned staff thing is not in self._staff_thing_list, create staff
                    # connect & register staff thing to middleware
                    self._send_TM_REGISTER(staff_thing)
                    self._send_TM_ALIVE(staff_thing)

                # update the scan cycle to half the minimum alive cycle of all staff thing
                self._scan_cycle = (
                    (min([staff_thing.get_alive_cycle() for staff_thing in self._staff_thing_list]) / 2.5)
                    if len(self._staff_thing_list) > 0
                    else self._scan_cycle
                )
                self._last_scan_time = get_current_time()
        except Exception as e:
            stop_event.set()
            print_error(e)
            return False

    # ====================================================================================================================
    #  _                        _  _        ___  ___ _____  _____  _____
    # | |                      | || |       |  \/  ||  _  ||_   _||_   _|
    # | |__    __ _  _ __    __| || |  ___  | .  . || | | |  | |    | |    _ __ ___    ___  ___  ___   __ _   __ _   ___
    # | '_ \  / _` || '_ \  / _` || | / _ \ | |\/| || | | |  | |    | |   | '_ ` _ \  / _ \/ __|/ __| / _` | / _` | / _ \
    # | | | || (_| || | | || (_| || ||  __/ | |  | |\ \/' /  | |    | |   | | | | | ||  __/\__ \\__ \| (_| || (_| ||  __/
    # |_| |_| \__,_||_| |_| \__,_||_| \___| \_|  |_/ \_/\_\  \_/    \_/   |_| |_| |_| \___||___/|___/ \__,_| \__, | \___|
    #                                                                                                         __/ |
    #                                                                                                        |___/
    # ====================================================================================================================

    # ===========================
    #            _____   _____
    #     /\    |  __ \ |_   _|
    #    /  \   | |__) |  | |
    #   / /\ \  |  ___/   | |
    #  / ____ \ | |      _| |_
    # /_/    \_\|_|     |_____|
    # ===========================

    @abstractmethod
    def _scan_staff_thing(self, timeout: float) -> List[dict]:
        '''
        지속적으로 staff thing을 발견하여 정보를 수집하여 반환하는 함수.
        timeout을 지정하여 한 번 staff thing을 검색하는데 소요될 시간을 지정할 수 있다.

        Args:
            timeout (float): staff thing을 검색하는데 소요될 시간

        Returns:
            List[dict]: staff thing의 정보를 담고 있는 리스트
        '''
        pass

    @abstractmethod
    def _create_staff(self, staff_thing_info: dict) -> Union[MXStaffThing, None]:
        '''
        _scan_staff_thing() 함수를 통해 수집된 staff thing 정보를 바탕으로 staff thing을 생성하는 함수.
        만약 스캔하는 것만으로 완벽한 staff thing의 정보를 수집할 수 없다면, staff thing의 register 메시지를 받아 처리하는
        _handle_REGISTER_staff_message() 함수에서 staff thing을 self._staff_thing_list에서 찾아 정보를 추가할 수 있다.

        Args:
            staff_thing_info (dict): staff thing의 정보를 담고 있는 딕셔너리

        Returns:
            staff_thing(MXStaffThing): 생성한 staff thing 인스턴스
        '''
        pass

    # ===============
    # ___  ___ _____
    # |  \/  ||_   _|
    # | .  . |  | |
    # | |\/| |  | |
    # | |  | |  | |
    # \_|  |_/  \_/
    # ===============

    @override
    def _handle_MT_RESULT_REGISTER(self, msg: mqtt.MQTTMessage) -> MXErrorCode:
        register_result_msg = MXRegisterResultMessage(msg)
        error = register_result_msg.error

        if register_result_msg.thing_name != self._name:
            target_thing: MXStaffThing = self._register_waiting_staff_thing_queue.get(timeout=5)
        else:
            target_thing = self

        if target_thing.get_name() != register_result_msg.thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Wrong payload arrive... {target_thing.get_name()} should be arrive, not {register_result_msg.thing_name}'
            )
            return error

        if error in [MXErrorCode.NO_ERROR, MXErrorCode.DUPLICATE]:
            MXLOG_DEBUG(f'{PrintTag.GOOD if error == MXErrorCode.NO_ERROR else PrintTag.DUP} Thing {target_thing.get_name()} register success!')
            target_thing.set_middleware_name(register_result_msg.middleware_name)
            target_thing.set_registered(True)
            self._subscribe_service_topic_list(target_thing)

            if isinstance(target_thing, MXStaffThing):
                self._staff_thing_list.append(target_thing)

            return error
        elif error == MXErrorCode.FAIL:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Register failed... Thing {target_thing.get_name()} register packet is not valid. error code: {register_result_msg.error}',
                'red',
            )
            return error
        else:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Register failed... Unexpected error occurred!!! error code: {register_result_msg.error}',
                'red',
            )
            return error

    @override
    def _handle_MT_RESULT_UNREGISTER(self, msg: mqtt.MQTTMessage) -> MXErrorCode:
        unregister_result_msg = MXUnregisterResultMessage(msg)
        error = unregister_result_msg.error

        if unregister_result_msg.thing_name != self._name:
            target_thing: MXStaffThing = self._unregister_waiting_staff_thing_queue.get(timeout=5)
        else:
            target_thing = self

        if target_thing.get_name() != unregister_result_msg.thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Wrong payload arrive... {target_thing.get_name()} should be arrive, not {unregister_result_msg.thing_name}',
                'red',
            )
            return error

        if error in [MXErrorCode.NO_ERROR, MXErrorCode.DUPLICATE]:
            MXLOG_DEBUG(f'{PrintTag.GOOD if error == MXErrorCode.NO_ERROR else PrintTag.DUP} Thing {target_thing.get_name()} register success!')
            target_thing.set_registered(False)
            self._unsubscribe_all_topic_list(target_thing)
            return error
        elif error == MXErrorCode.FAIL:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Unregister failed... Thing {target_thing.get_name()} unregister packet is not valid. error code: {unregister_result_msg.error}',
                'red',
            )
            return error
        else:
            MXLOG_DEBUG(
                f'{PrintTag.ERROR} Unregister failed... Unexpected error occurred!!! error code: {unregister_result_msg.error}',
                'red',
            )
            return error

    @override
    def _handle_MT_RESULT_BINARY_VALUE(self, msg: mqtt.MQTTMessage) -> bool:
        execute_msg = MXBinaryValueResultMessage(msg)
        staff_thing = self._get_staff_thing_by_name(execute_msg.thing_name)

        if staff_thing.get_name() != execute_msg.thing_name:
            MXLOG_DEBUG(
                f'[{get_current_function_name()}] Wrong payload arrive... {staff_thing.get_name()} should be arrive, not {execute_msg.thing_name}',
                'red',
            )
            return False

        for value in staff_thing.get_value_list():
            if value.get_name() == execute_msg.value_name and value.get_type() == MXType.BINARY:
                value.set_binary_sending(False)
                return True
        else:
            MXLOG_DEBUG(f'[{get_current_function_name()}] Value {execute_msg.value_name} does not exist!!!', 'red')
            return False

    @override
    def _handle_MT_EXECUTE(self, msg: mqtt.MQTTMessage) -> bool:
        execute_msg = MXExecuteMessage(msg)

        if execute_msg.thing_name != self._name:
            target_thing = self._get_staff_thing_by_name(execute_msg.thing_name)
        else:
            target_thing = self

        target_function = target_thing._get_function(execute_msg.function_name)
        if target_function:
            execute_thread = target_function.start_execute_thread(execute_msg)
            return True
        else:
            MXLOG_DEBUG('function not exist', 'red')
            return False

    # ===============
    #  _____ ___  ___
    # |_   _||  \/  |
    #   | |  | .  . |
    #   | |  | |\/| |
    #   | |  | |  | |
    #   \_/  \_|  |_/
    # ===============

    @override
    def _send_TM_REGISTER(self, thing: MXThing) -> None:
        if isinstance(thing, MXStaffThing):
            self._staff_thing_list.append(thing)
            thing._last_alive_time = get_current_time()
            self._register_waiting_staff_thing_queue.put(thing)

        self._subscribe_init_topic_list(thing)
        register_msg = thing.generate_register_message()
        if not register_msg:
            raise Exception('TM_REGISTER packet error')

        register_mqtt_msg = register_msg.mqtt_message()
        self._publish_queue.put(register_mqtt_msg)

    @override
    def _send_TM_UNREGISTER(self, thing: MXThing):
        if isinstance(thing, MXStaffThing):
            self._staff_thing_list.remove(thing)
            self._unregister_waiting_staff_thing_queue.put(thing)

        unregister_msg = thing.generate_unregister_message()
        if not unregister_msg:
            raise Exception('TM_UNREGISTER packet error')

        unregister_mqtt_msg = unregister_msg.mqtt_message()
        self._publish_queue.put(unregister_mqtt_msg)

    @override
    def _send_TM_ALIVE(self, staff_thing: MXThing):
        staff_thing._last_alive_time = get_current_time()
        alive_msg = staff_thing.generate_alive_message()
        alive_mqtt_msg = alive_msg.mqtt_message()
        self._publish_queue.put(alive_mqtt_msg)
        staff_thing.set_last_alive_time(get_current_time())

    @override
    def _send_TM_VALUE_PUBLISH(self, staff_thing: MXThing, value: MXValue) -> None:
        value_publish_msg = staff_thing.generate_value_publish_message(value=value)
        value_publish_mqtt_msg = value_publish_msg.mqtt_message()

        if value.get_type() == MXType.BINARY:
            value.set_binary_sending(True)
        self._publish_queue.put(value_publish_mqtt_msg)

    @override
    def _send_TM_VALUE_PUBLISH(self, staff_thing: MXThing, value: MXValue) -> None:
        value_publish_msg = staff_thing.generate_value_publish_message(value=value)
        value_publish_mqtt_msg = value_publish_msg.mqtt_message()

        if value.get_type() == MXType.BINARY:
            value.set_binary_sending(True)
        self._publish_queue.put(value_publish_mqtt_msg)

    # ========================
    #         _    _  _
    #        | |  (_)| |
    #  _   _ | |_  _ | | ___
    # | | | || __|| || |/ __|
    # | |_| || |_ | || |\__ \
    #  \__,_| \__||_||_||___/
    # ========================

    @override
    def _register(self, thing: MXThing):
        self._send_TM_REGISTER(thing)

    @override
    def _subscribe_init_topic_list(self, staff_thing: MXThing) -> None:
        topic_list = [
            MXProtocolType.Base.MT_RESULT_REGISTER.value % staff_thing.get_name(),
            MXProtocolType.Base.MT_RESULT_UNREGISTER.value % staff_thing.get_name(),
            MXProtocolType.Base.MT_RESULT_BINARY_VALUE.value % staff_thing.get_name(),
        ]

        for topic in topic_list:
            self._subscribe(topic)

    @override
    def _subscribe_service_topic_list(self, staff_thing: MXThing):
        topic_list = []

        for function in staff_thing.get_function_list():
            topic_list += [
                MXProtocolType.Base.MT_EXECUTE.value % (function.get_name(), staff_thing.get_name(), '+', '+'),
                (MXProtocolType.Base.MT_EXECUTE.value % (function.get_name(), staff_thing.get_name(), '', '')).rstrip('/'),
            ]

        for topic in topic_list:
            self._subscribe(topic)

    @override
    def _unsubscribe_all_topic_list(self, staff_thing: MXThing):
        # whenever _unsubscribe function execute, it remove target topic from self._subscribed_topic_set
        # so it need to iterate with copy of self._subscribed_topic_set
        topic_list = list(staff_thing.get_subscribed_topic_set())

        for topic in topic_list:
            self._unsubscribe(topic)

    @override
    def dict(self) -> dict:
        return {
            "name": self._name,
            "alive_cycle": self._alive_cycle,
            "is_super": self._is_super,
            "is_parallel": self._is_parallel,
            "is_manager": True,
            "is_matter": None,
            "values": [value.dict() for value in self._value_list],
            "functions": [function.dict() for function in self._function_list],
        }

    def generate_staff_thing_id(self) -> str:
        return str(uuid.uuid4())

    def _get_staff_thing_by_name(self, staff_name: str, staff_thing_pool: List[MXStaffThing] = None) -> MXStaffThing:
        if staff_thing_pool is None:
            staff_thing_pool = self._staff_thing_list

        for staff_thing in staff_thing_pool:
            if staff_thing.get_name() == staff_name:
                return staff_thing

        return False

    def _get_staff_thing_by_staff_thing_id(self, staff_thing_id: str, staff_thing_pool: List[MXStaffThing] = None) -> MXStaffThing:
        if staff_thing_pool is None:
            staff_thing_pool = self._staff_thing_list

        for staff_thing in staff_thing_pool:
            if staff_thing._staff_thing_id == staff_thing_id:
                return staff_thing

        return False



from big_thing_py.utils.log_util import *
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning


requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


class RequestMethod(Enum):
    GET = 0
    POST = 1
    PUT = 2
    DELETE = 3


def API_response_check(res: requests.Response):
    if res.status_code not in [200, 204]:
        return False
    else:
        return res


def API_request(
    url,
    method: RequestMethod = RequestMethod.GET,
    header: str = '',
    body: str = '',
    verify: bool = False,
    timeout: float = 10,
) -> dict:
    try:
        if method == RequestMethod.GET:
            res = requests.get(url, headers=header, verify=verify, timeout=(1, timeout))
            if API_response_check(res):
                if res.status_code == 200:
                    data = res.json()
                elif res.status_code == 204:
                    data = {}
                return data
            else:
                return False
        elif method == RequestMethod.POST:
            res = requests.post(url, headers=header, data=body, verify=verify, timeout=(1, timeout))
            if API_response_check(res):
                return res
            else:
                return False
        elif method == RequestMethod.PUT:
            res = requests.put(url, headers=header, data=body, verify=verify, timeout=(1, timeout))
            if API_response_check(res):
                return res
            else:
                return False
        elif method == RequestMethod.DELETE:
            MXLOG_DEBUG('Not implement yet')
        else:
            MXLOG_DEBUG(f'[decode_MQTT_message] Unexpected request!!!', 'red')
    except Exception as e:
        MXLOG_DEBUG(f'Failed to request API: {e}', 'red')
        return False


if __name__ == '__main__':
    res = API_request('https://www.google.com')
    print(res)



from big_thing_py.utils import *
from enum import Enum, auto


class MXManagerMode(Enum):
    UNDEFINED = 'UNDEFINED'
    JOIN = 'JOIN'
    SPLIT = 'SPLIT'

    @classmethod
    def get(cls, name: str) -> 'MXManagerMode':
        try:
            return cls[name.upper()]
        except Exception:
            return cls.UNDEFINED


class MXStaffThingInfo:
    def __init__(self, staff_thing_id: str) -> None:
        self.staff_thing_id = staff_thing_id


class StaffRegisterResult:
    def __init__(self, staff_thing_name: str, staff_thing_id: str, assigned_staff_thing_id: str) -> None:
        self.staff_thing_name = staff_thing_name
        self.staff_thing_id = staff_thing_id
        self.assigned_staff_thing_id = assigned_staff_thing_id
